#include <algorithm>
#include "Utilities.h"

using std::string;
using std::vector;
using std::domain_error;
using std::max;
using std::cout;
using std::endl;

/*******************************************************************************	
*									       *	
*  Compute the final result when Fixed Order Calculation is requested          *
*									       *	
*******************************************************************************/

double ComputeFinalResultFO(PartonicFlux Flux,
			    FixedOrder FO, 
			    double rho, 
			    int precision)
{
  return ConvolutionFO(Flux,FO, rho, precision);
}

double f_FO (double x, 
	     void * par) 
{
  FOcomponents * p = (FOcomponents*) par;
  double f = ( 
	      (p->Flux).fluxqqbar(x)      * (p->FO).FOqqbar((p->rho)/x)     +
	      (p->Flux).fluxgg(x)         * (p->FO).FOgg((p->rho)/x)        +
	      (p->Flux).fluxgq(x)         * (p->FO).FOgq((p->rho)/x)        +
	      (p->Flux).fluxqq(x)         * (p->FO).FOqq((p->rho)/x)        +
	      (p->Flux).fluxqqprime(x)    * (p->FO).FOqqprime((p->rho)/x)   +
	      (p->Flux).fluxqqbarprime(x) * (p->FO).FOqqbarprime((p->rho)/x)
	       )/x;
  return f;
}

double ConvolutionFO(PartonicFlux Flux, 
		     FixedOrder FO, 
		     double rho, 
		     int precision)
{
  FOcomponents temp_p = {Flux, FO , rho};
  gsl_function F;
  F.function = f_FO;
  F.params = &temp_p;
  //------------------- setup for the integration routine (not meant to be modified):
  int prec = precision;
  double a = rho;
  double b = 1.0;
  double epsabs = 0.0;
  double epsrel = 1/pow(10,prec);
  size_t limit = 1000;
  int key = 6;
  //-------------------------------
  double result, abserr;
  gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);	
  gsl_integration_qag (&F, a, b, epsabs, epsrel,limit, key, w, &result, &abserr); 
  gsl_integration_workspace_free (w);
  
  if (abserr/result > epsrel) 
    {
      cout << "Warning (Fixed Order cross-section convolution): relative error reached " 
	   << abserr/result 
	   << endl;
    }
  return result;
}

/*******************************************************************************	
*									       *	
*        Compute the final result for mixed F.O. with Resummation              *
*									       *	
*******************************************************************************/

double ComputeFinalResultRES(PartonicFlux Flux,
			     SubtrFlux SFlux, 
			     FixedOrder FO, 
			     Resummation Res, 
			     double rho, 
			     int precision)
{	
  double f =  ConvolutionRES(Flux, SFlux, Res, rho, precision)
    + Res.resumqqbarSubFlux(rho) 
    + Res.resumggSubFlux(rho);
  return f;
}

double ConvolutionRES(PartonicFlux Flux, 
		      SubtrFlux SFlux, 
		      Resummation Res, 
		      double rho, 
		      int precision)
{
  REScomponents temp_p = {Flux, SFlux, Res, rho};
  gsl_function F;
  F.function = f_RES;
  F.params = &temp_p;
  //------------------- setup for the integration routine (not meant to be modified):
  int prec = precision;
  double a = rho;
  double b = 1.0;
  double epsabs = 0.0;
  double epsrel = 1/pow(10,prec);
  size_t limit = 1000;
  int key = 6;
  //-------------------------------
  double result, abserr;
  gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);	
  gsl_integration_qag (&F, a, b, epsabs, epsrel,limit, key, w, &result, &abserr); 
  gsl_integration_workspace_free (w);
  if (abserr/result > epsrel) 
    {
      cout << "Warning (Resummed cross-section x-space convolution): relative error reached " 
	   << abserr/result 
	   << endl;
    }
  return result;
}

double f_RES (double x, void * par) 
{
  REScomponents * p = (REScomponents*) par;	
  double f = (   ((p->Flux).fluxqqbar(x) 
		  - (p->SFlux).subtrfluxqqbar(x) ) * (p->Res).resumqqbar((p->rho)/x) 
		 + ((p->Flux).fluxgg(x)    
		    - (p->SFlux).subtrfluxgg(x) )  * (p->Res).resumgg((p->rho)/x)   
		 )/x;
  return f;
}


/*******************************************************************************	
*									       *	
*        If Fact. and Ren. scales pass the scale variations criterion.         *
*									       *	
*******************************************************************************/

bool ScalesPass(string RestrictedScaleVariation, 
		double F, 
		double R)
{// At this point RestrictedScaleVariation is either "NO" or a "double >= 1.0"
  if (RestrictedScaleVariation=="NO") return 1;
  else {
    double x = atof(RestrictedScaleVariation.c_str());
    if (F/R <= x && F/R >= 1/x) return 1;
    else return 0;
  }
}

/*******************************************************************************	
*									       *	
*        Converts the options OrderRES and OrderFO to integers:                *
*									       *	
*******************************************************************************/

int IntegerForm(string s)
{
  if      (s=="LO"   || s=="LL")   {return 1;}
  else if (s=="NLO"  || s=="NLL")  {return 2;}
  else if (s=="NNLO" || s=="NNLL") {return 3;}
  else 
    {
      cout << "Warning: unknown option for OrderRES or OrderFO (LO/LL,NLO/NLL,NNLO/NNLL expected)." 
	   << endl;
      return 0;
    }
}

/*******************************************************************************	
*									       *	
*			             Computing PDF uncertainty:                *
*            The asymmetric prescription used by MSTW and others.              *
*            (see Cacciari et al. http://arxiv.org/abs/0804.2800)              *
*									       *	
*******************************************************************************/	


void deltaPDFasymmetric(const vector<double>& result, 
			double central, 
			double& smax, 
			double& smin)
{
  
  if (LHAPDF::numberPDF()%2 != 0) 
    {
      throw domain_error("The number of pdf eigenvectors is not even. Don't know how to proceed.");
    }
  int Imax = LHAPDF::numberPDF()/2; //int division OK; numberPDF is even
  
  double odd, even;
  double tp,tm;
  vector<double> plus,minus;
  
  for (int i = 1; i<=Imax; ++i ) 
    {
      odd  = result[2*i-1] - central;
      even = result[2*i] - central;
      
      tp = max( max(even,odd) ,0.0);
      tm = max( max(-even,-odd) ,0.0);
      
      plus.push_back(  tp );
      minus.push_back( tm );
    }
  
  double deltaP=0.0;
  for(vector<double>::iterator it = plus.begin(); it != plus.end(); ++it) 
    {
      deltaP += pow( *it ,2);
    }
  double deltaM=0.0;
  for(vector<double>::iterator it = minus.begin(); it != minus.end(); ++it) 
    {
      deltaM += pow( *it ,2);
    }
  
  smax = central + sqrt(deltaP);
  smin = central - sqrt(deltaM);
}

/*******************************************************************************	
*									       *	
*			             Computing PDF uncertainty:                *
*              The prescription used by the NNPDF family of sets.              *
*									       *	
*******************************************************************************/	

void deltaPDFnnpdf(const std::vector<double>& result, 
		   double& smax, 
		   double& smin)
{
  // The first element of 'result' has to be excluded;
  // it is an approximation to 'mean' to be derived below.	
  
  int size = result.size()-1;
  double ar[size];
  for(int i = 0; i != size; ++i)
    {
      ar[i] = result[i+1];
    }
  
  double mean = gsl_stats_mean (ar, 1, size);
  double sd = gsl_stats_sd_m (ar, 1, size, mean);
  
  smax = mean + sd;
  smin = mean - sd;
}

/*******************************************************************************	
 *									       *	
 *			             Computing PDF uncertainty:                *
 *   The symmetric prescription used by the Alekhin family of sets             *
 *                       (including ABKM09 and ABM11)                          *
 *									       *	
 *******************************************************************************/	

void deltaPDFsymmetric(const std::vector<double>& result, 
		       double& smax, 
		       double& smin)
{	
  int size = result.size();
  double deltasq = 0;
  for(int i = 0; i != size; ++i)
    {
      deltasq += pow(result[0]-result[i],2);
    }
  double delta = sqrt(deltasq);
  smax = result[0] + delta;
  smin = result[0] - delta;
}
